from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, ForeignKey, UniqueConstraint
from sqlalchemy import Integer, String, DateTime, Numeric, Boolean
from sqlalchemy.orm import relationship, Session
from geoalchemy2 import Geometry

from datetime import datetime as dt


class Base(object):
    # overwrite the default values
    __prepopulate__ = []

    @classmethod
    def defaults(cls, session: Session):
        instances = [cls(**d) for d in cls.__prepopulate__]

        try:
            session.add_all(instances)
            session.commit()
        except Exception as e:
            session.rollback()
            print('Populating defaults errored\n%s' % str(e))

# augment the sqlalchemy Base class with defaults classmethod
Base = declarative_base(cls=Base)

# ---------------------------------------------------------------------
# Database Models
# ---------------------------------------------------------------------
class DetailAssociation(Base):
    __tablename__ = 'nm_metadata_details'

    meta_id = Column(Integer, ForeignKey('metadata.id'), primary_key=True)
    detail_id = Column(Integer, ForeignKey('details.id'), primary_key=True)

class Metadata(Base):
    __tablename__ = 'metadata'

    # columns
    id = Column(Integer, primary_key=True)
    device_id = Column(String, nullable=False)
    sensor_id = Column(Integer, ForeignKey('sensors.id'))
    term_id = Column(Integer, ForeignKey('terms.id'))
    location = Column(Geometry('POINT', srid=4326), nullable=False)
    description = Column(String, nullable=True)

    # relationships
    term = relationship('Term', back_populates='data')
    details = relationship('Detail', secondary='nm_metadata_details', back_populates='meta', cascade='all,delete')

    def __init__(self, **kwargs):
        # extract the column names of this model
        column_names = [col.name for col in Metadata.__table__.columns]

        # split the input into core and details
        in_names = {k: v for k, v in kwargs.items() if k in column_names}
        out_names = {k: v for k, v in kwargs.items() if k not in column_names}
        
        # run Base class __init__ method
        super(Metadata, self).__init__(**in_names)

        # add details
        self.set_details(out_names)

    def get_details(self):
        details = dict()
        for detail in self.details:
            detail.to_dict(add_to=details)
        return details

    def set_details(self, details: dict):
        for k,v in details.items():
            self.details.append(Detail(k, v))


class Variable(Base):
    __tablename__ = 'variables'
    __prepopulate__ = [
        dict(name='temperature', unit='deg. Celsius'),
        dict(name='light', unit='intensity', comment='Ambient light intensity (can be converted to lux).')
    ]

    id = Column(Integer, primary_key=True)
    name = Column(String(64), nullable=False, unique=True)
    unit = Column(String(30), nullable=False)
    comment = Column(String)


class Sensor(Base):
    __tablename__ = 'sensors'
    
    id = Column(Integer, primary_key=True)
    name = Column(String(), nullable=False)


class Term(Base):
    __tablename__ = 'terms'

    # columns
    id = Column(Integer, primary_key=True)
    short = Column(String(8), nullable=False)
    full_name = Column(String(), nullable=False)
    start_date = Column(DateTime, nullable=False)
    end_date = Column(DateTime, nullable=False)

    # relationships
    data = relationship('Metadata', back_populates='term')

    @classmethod
    def defaults(cls, session: Session):
        """
        Load some default values
        """
        # build the default values
        cls.__prepopulate__ = []
        for y in range(16, 40):
            y2 = y + 1
            cls.__prepopulate__.extend([
                dict(short='WT%d' % y, full_name='Winterterm 20%d/20%d' % (y, y2), start_date=dt(2000 + y, 10, 1), end_date=dt(2000 + y2, 3, 31)),
                dict(short='ST%d' % y2, full_name='Summerterm 20%d' % y2, start_date=dt(2000 + y2, 4, 1), end_date=dt(2000 + y2, 9, 30)),
            ])
        
        # super into Base class
        super(Term, cls).defaults(session=session)

    def __str__(self):
        return self.short


class Quality(Base):
    __tablename__ = 'quality'

    id = Column(Integer, primary_key=True)
    short = Column(String(20), nullable=False, unique=True)
    name = Column(String)
    comment = Column(String)


class Data(Base):
    __tablename__ = 'data'

    meta_id = Column(Integer, ForeignKey('metadata.id'), primary_key=True)
    variable_id = Column(Integer, ForeignKey('variables.id'), primary_key=True)
    tstamp = Column(DateTime, primary_key=True)
    value = Column(Numeric, nullable=False)
    quality_flag_id = Column(Integer, ForeignKey('quality.id'), nullable=False)


class RawData(Base):
    __tablename__ = 'raw_data'
#    __table_args__ = (
#        UniqueConstraint('meta_id', 'variable_id', 'tstamp', name='uq_raw_data')
#    )

    meta_id = Column(Integer, ForeignKey('metadata.id'), primary_key=True)
    variable_id = Column(Integer, ForeignKey('variables.id'), primary_key=True)
    tstamp = Column(DateTime, primary_key=True)
    value = Column(Numeric, nullable=False)


class Detail(Base):
    __tablename__ = 'details'

    # columns
    id = Column(Integer, primary_key=True)
    key = Column(String, nullable=False)
    str_value = Column(String, nullable=True)
    int_value = Column(Integer, nullable=True)
    float_value = Column(Numeric, nullable=True)
    bool_value = Column(Boolean, nullable=True)

    # relationships
    meta = relationship('Metadata', secondary='nm_metadata_details', back_populates='details')

    def __init__(self, key, value):
        self.key = key
        if isinstance(value, str):
            self.str_value = value
        elif isinstance(value, int):
            self.int_value = value
        elif isinstance(value, float):
            self.float_value = value
        elif isinstance(value, bool):
            self.bool_value = value
        else:
            self.str_value = str(value)

    def to_dict(self, add_to=None):
        if add_to is None:
            add_to = dict()
        for _t in ['str', 'int', 'float', 'bool']:
            value = getattr(self, '%s_value' % _t)
            if value is not None:
                add_to[self.key] = value
        return add_to


# -----------------------------------------------------------
# Some extra models
# -----------------------------------------------------------
class SpaceRaw(Base):
    __tablename__ = 'space_raw'

    id = Column(Integer, primary_key=True)
    company_name = Column(String(65))
    location = Column(String)
    datum = Column(DateTime)
    detail = Column(String)
    status_rocket = Column(String)
    status_mission = Column(String)



import pandas as pd 
import progressbar
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from hydenv.util import env
from hydenv.models import SpaceRaw


class HydenvSpaceExamples:
    """
    Space Missions Example Loader.\n
    Loads Space Missions example data from Github and load it into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", overwrite_url=None):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

        # set some URLs
        if overwrite_url is None:
            self._data_url_or_path = 'https://raw.githubusercontent.com/data-hydenv/data/master/extra/sample/space/space.csv'
        else:
            self._data_url_or_path = overwrite_url

    def run(self, quiet=True):
        """
        """
        # download dataset
        df = pd.read_csv(self._data_url_or_path)

        # drop the first two columns
        df.drop(df.columns[:2], axis=1, inplace=True)

        # rename columns
        df.columns = [_.lower().replace(' ', '_')  for _ in df.columns]
        df.drop('_rocket', axis=1, inplace=True)

        # convert datum to UTC 
        df['datum'] = pd.to_datetime(df.datum, utc=True)

        # rename index to 'id'
        df.index.name = 'id'
        
        if not quiet:
            print('Uploading Space Missions example data...')
            bar = progressbar.ProgressBar(max_value=len(df), redirect_stdout=True)
        else:
            df.to_sql('space_raw', self.engine, if_exists='replace')
            return

        # upload missions
        for i, space in df.iterrows():
            mission = SpaceRaw(**space)

            # add
            try:
                self.session.add(mission)
                self.session.commit()
            except:
                self.session.rollback()
                if not quiet:
                    print('[%d] Error.' % i)
            
            if not quiet:
                bar.update(i + 1)
        
        if not quiet:
            print('Done.')



        


from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from overpy import Overpass
from progressbar import ProgressBar

from hydenv.util import env
from hydenv import models

# set some helpful overpass API queries

BY_AREA = """
area[boundary~"administrative"][name="{boundary_name}"];
node(area)[{attribute}{value}];out;
"""


class HydenvOSMExamples:
    """
    OSM Example Loader.\n
    Loads Open Street Map example either as persisted dumps from Github, or 
    dynamically direclty from the OSM database using Overpass API. The data 
    is then loaded into the given database. The database has to be installed 
    and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", overpass_url=None):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

        # set the Overpass API
        self.api = Overpass(url=overpass_url)

    def run(self, boundary, attribute, value=None, dry=False, save=None, quiet=True):
        # first build the query
        value = '=%s' % value if value is not None else ''
        query = BY_AREA.format(boundary_name=boundary, attribute=attribute, value=value)

        # if dry run, do not reach out to the Overpass server
        if dry:
            if not quiet:
                return query
            else:
                print('Overpass QL:\n%s' % query)
                return
        
        # get the data
        if not quiet:
            print('Connecting %s...' % self.api.url)
        try:
            result = self.api.query(query)
            nodes = result.get_nodes()
        except Exception as e:
            print('The following error was caused during Overpass connection:\n%s' % str(e))
            return
        
        # inform and create the progressbar
        if not quiet:
            print('Uploading...')
            bar = ProgressBar(max_value=len(nodes), redirect_stdout=True)
        
        for i, node in enumerate(nodes):
            wkt = 'SRID=4326;POINT (%f %f)' % (float(node.lon), float(node.lat))
            osm = models.OSMNode(
                id=node.id, 
                geom=wkt, 
                name=node.tags.get('name'),
                node_type=value
            )
            for k,v in node.tags.items():
                osm.raw_tags.append(models.OSMTag(key=k, value=v))

            # upload
            try:
                self.session.add(osm)
                self.session.commit()
            except Exception as e:
                self.session.rollback()
                if not quiet:
                    print('[%d]: %s' % (osm.id, str(e)))
                continue
            if not quiet:
                bar.update(i + 1)
        if not quiet:
            print('Done')
        

    def energiewende(self, boundary="Baden-Württemberg", quiet=True):
        # run for 
        if not quiet:
            print('Loading all Gas stations in %s' % boundary)
        self.run(boundary, 'amenity', 'fuel', quiet=quiet)

        if not quiet:
            print('Loading all charging stations in %s' % boundary)
        self.run(boundary, 'amenity', 'charging_station', quiet=quiet)

        if not quiet:
            print('Energiewende loaded.')



from hydenv.util import env


class HydenvEarthquakeExamples:
    """
    Eaarthquake Example Loader.\n
    Loads earthquake example data from Github (or other sources) and loads them into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)

        # set some URLs
        self._github_url = 'https://raw.githubusercontent.com/data-hydenv/data/master/extra/spatial/earthquake/earthquakes.csv'

    def run(self, normalize=False):
        """
        """
        raise NotImplementedError



import pandas as pd
import os
import ntpath
import re
from string import ascii_lowercase
from random import choice
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from datetime import datetime as dt
from progressbar import ProgressBar

from hydenv.util import env
from hydenv.measurements import HydenvMeasurements
from hydenv.models import Metadata


def read_file(fname, txtfmt=False):
	if txtfmt:
		# HOBO TXT format
		df = pd.read_csv(fname, skiprows=2, header=None, sep='\s+', thousands=',', na_values='Logged')
		df['tstamp'] = [dt.strptime(row[1] + ' ' + row[2], '%d-%m-%y %H:%M:%S') for i, row in df.iterrows()]
		df.drop([0,1,2], axis=1, inplace=True)
		df.columns = ['temperature', 'light', 'tstamp']
	else:
		# HOBO CSV format
		df = pd.read_csv(fname, skiprows=2, header=None, parse_dates=[1], usecols=[1,2,3])
		#df.drop(0, axis=1, inplace=True)
		df.dropna(inplace=True)
		df.columns = ['tstamp', 'temperature', 'light']

	# make sure temperature and light are floats
	df.temperature = df.temperature.astype(float)
	df.light = df.light.astype(float)

	# drop na
	df.dropna(how='any', axis=0, inplace=True)
	
	return df


class HydenvHoboImporter:
	"""
    Measurements management tool\n
    This is the main tool for CRUD operations on the database.
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
	def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        # substitute and save
		self.__connection = env.build_connection(connection=connection)
		self.engine = create_engine(self.__connection)
		Session = sessionmaker(bind=self.engine)
		self.session = Session()

	def metadata(self, url: str, term: str = None, if_exists='append'):
		"""
		Copy metadata from google tables.\n
		Make sure that the URL format is correct. Open the correct sheet and copy the url,
		then change the last parameter from 'edit#gid=...' to '/export?gid=...'
		:param url: URL to the table
		:param term: the short name of the term, like WS17
		:param if_exists: How to handle existing table. Can be one of:
			- append (default) - append data to the table
			- fail - raise an error and exit
			- replace - drop the table and recreate
		"""
		# use correct URL
		if 'edit#gid=' in url:
			url = url.replace('edit#', 'export?')
		if not 'format=csv' in url:
			url += '&format=csv'

		# download
		df = pd.read_csv(url)

#		imp = df[['hobo_id', 'description']].copy()
		df['location'] = df[['longitude', 'latitude']].apply(lambda r: 'SRID=4326;POINT (%s %s)' % (r[0], r[1]), axis=1)
		df.drop(['longitude', 'latitude'], axis=1, inplace=True)
		df.dropna(axis=1, how='all', inplace=True)
#		imp.columns = ['device_id', 'description', 'location']
		df.rename({'hobo_id': 'device_id'}, axis=1, inplace=True)

		# check if the sensor 'hobo' exists
		cli = HydenvMeasurements(self.__connection)
		hobo = cli.read('Sensor', return_query=True, name='hobo').first()
		semester = cli.read('Term', return_query=True, short=term).first()
		
		# if there is no HOBO, create one
		if hobo is None:
			hobo = cli.create('Sensor', name='hobo')
		
		# if semester is given, add
		if semester is not None:
			df['term_id'] = semester.id
		
		# add sensor info
		df['sensor_id'] = hobo.id

		# build an engine
#		engine = create_engine(self.__connection)

		# upload
		#imp.to_sql('metadata', engine, if_exists=if_exists, index=False)
#		metas = []
#		for d in df.to_dict('records'):
#			meta = Metadata(**d)
#			meta.set_details({k:v for k,v in d.items() if k not in ignores})
#			metas.append(meta)
		try:
			self.session.add_all([Metadata(**d) for d in df.to_dict('records')])
			self.session.commit()
		except Exception as e:
			self.session.rollback()
			raise e

	def upload(self, filename: str, meta_id: int, variable=None):
		"""
		Upload an Hobo file to the database. 
		:param filename: Path to the file for upload
		:param meta_id: Database ID of your device
		"""
		# check which variable should be uploaded
		if variable is None:
			self.upload(filename, meta_id, 'temperature')
			self.upload(filename, meta_id, 'light')
			return
		
		# load the file
		txtfmt = filename.endswith('.txt')
		try:
			data = read_file(filename, txtfmt=txtfmt)
		except Exception as e:
			print("Parsing file '%s' was not successfull.\nDo not edit the files by hand!\nError: %s " % (filename, str(e)))
			return

#		# get the hobo id
#		hobo_id = os.path.split(filename)[-1].split('.')[0]

		# search the metadata
		cli = HydenvMeasurements(connection=self.__connection)
		#meta = cli.read('Metadata', id=meta_id)

		# check if devices exists, if yes check timestamps

		# get the ovariable ids
		var_t = cli.read('Variable', name=variable, return_query=True).one()

		# upload temperature
		temp = data[['tstamp', variable]].copy()
		temp.columns = ['tstamp', 'value']
		temp['meta_id'] = meta_id
		temp['variable_id'] = var_t.id

		# create database engine
		engine = create_engine(self.__connection)
		
		# create a temporary table name
		name = ''.join([choice(ascii_lowercase) for _ in range(16)])

		# upload to temporary table
		temp.to_sql(name, engine, index=False)
		engine.execute('INSERT INTO raw_data (meta_id, variable_id, tstamp, value) SELECT meta_id, variable_id, tstamp, value FROM %s ON CONFLICT (meta_id, variable_id, tstamp) DO NOTHING' % name)
		engine.execute('UPDATE raw_data r SET value=t.value FROM %s t WHERE r.meta_id=t.meta_id AND r.tstamp=t.tstamp' % name)

		# delete the temp table
		engine.execute("DROP TABLE %s" % name)

		# return the number of rows
		return len(temp)

	def folder(self, path='.', match=r'[0-9]+\.(txt|csv)', term: str = None, quiet=True):
		"""
		Upload the whole folder content given at path.\n
		Path defaults to the current location, but can be replaced.
		The default match pattern is '[0-9]+\.(txt|csv)'.
		:param path: Location to search, defaults to  './'
		:param match: Regular expression to match all data files
		:param term: If applicable, add the term for the data like: WS17
		"""
		flist = [os.path.join(path, f) for f in os.listdir(path) if re.match(match, f)]
		
		# load the term if any
		cli = HydenvMeasurements(self.__connection)
		if term is not None:
			semester = cli.read('Term', return_query=True, short=term).first()
			if semester is None:
				raise AttributeError("Term '%s' not found." % term)
			else: 
				term_id = semester.id
		else: 
			term_id = None

		# build a progressbar
		if not quiet:
			bar = ProgressBar(max_value=len(flist), redirect_stdout=True)
		
		# load all files
		for i, fname in enumerate(flist):
			try:
				device_id = ntpath.split(fname)[-1].split('.')[0]
				meta = cli.read('Metadata', return_query=True, device_id=device_id, term_id=term_id).first()
			except:
				print("File '%s' cannot be processed." % fname)
				continue
			
			# if there is no metadata, skip
			if meta is None:
				print('File %s references HOBO ID=%s, which is not found.' % (fname, device_id))
			else:
				# upload
				self.upload(filename=fname, meta_id=meta.id)
			if not quiet:
				bar.update(i + 1)			


class HydenvHoboImporterCli(HydenvHoboImporter):
	def upload(self, filename: str, meta_id: int, variable=None):
		# run the function
		count = super(HydenvHoboImporterCli, self).upload(filename=filename, meta_id=meta_id, variable=variable)
		if count is not None:
			print('[%s]: Uploaded %d %s measurements.' % (meta_id, count, variable))

	def folder(self, path='.', match=r'[0-9]+\.(txt|csv)', term: str = None):
		super(HydenvHoboImporterCli, self).folder(path=path, match=match, term=term, quiet=False)


if __name__=='__main__':
	import fire
	fire.Fire(HydenvHoboImporterCli)



"""
Examples
--------
This CLI loads example files from Github (or other sources) 
and loads them into the given database. 

Note
----
The database has to be installed and initialized before. 
Do that by hand or use the two CLI commands:

.. code-block:: bash

    python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
    python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv

"""
import progressbar
import os
import shutil
import zipfile
import requests
import io

from hydenv.util import env
from hydenv.scripts.hobo_import import HydenvHoboImporter


class HydenvHoboExamples:
    """
    Hobo Example Loader.\n
    Loads HOBO example data from Github (or other sources) and loads them into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}"):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)

        # set some URLs
        self._hobo_table_url = "https://docs.google.com/spreadsheets/d/1uvI1a_OOnCcFjUbHAO1N5nEWejnI5y5vCNt6eueHrOc/edit"
        self._hobo_data_url = "https://github.com/data-hydenv/data/archive/master.zip"
        self.__hobo_gid_map = {
            'WT17': "#gid=1594319231",
            'WT18': "#gid=1500046150",
            'WT19': '#gid=651397949',
            'WT20': "#gid=0"
        }
        self.__hobo_data_map = {
            'WT17': 'hobo/2017/',
            'WT18': "hobo/2018/",
            'WT19': 'hobo/2019/',
            'WT20': "hobo/2020/"
        }
    
    def run(self, terms='all', data_path='download', only=None, quiet=True):
        if only is None or only == 'metadata':
            # Get the metadata
            if not quiet:
                print('Uploading Metdata')
            self._upload_hobo(terms=terms, quiet=quiet)
            if not quiet:
                print('Done.')
        
        # download
        if only is None or only == 'data':
            try:
                if data_path == 'download':
                    if not quiet:
                        print('Downloading data repository. This may take a minute.')
                    self._download_data()
                    data_path = './data-master/'
                    if not quiet:
                        print('Done.')
                
                # upload
                if not quiet:
                    print('Start uploading. You can grab a coffee...')
                self._upload_data(data_path, terms=terms, quiet=quiet)

            except Exception as e:
                raise e
            finally:
                # delete the temp data 
                if os.path.exists(data_path):
                    shutil.rmtree(data_path)
        
        if not quiet:
                print('Done.')
        
    def _upload_hobo(self, terms='all', quiet=True):
        # if all years, are requested, build the list
        if terms == 'all':
            terms = ['WT17', 'WT18', 'WT19', 'WT20']
        if not isinstance(terms, list):
            terms = [terms]

        # build the cli
        cli = HydenvHoboImporter(connection=self.__connection)
        
        # build the upload bar
        if not quiet:
            bar = progressbar.ProgressBar(max_value=len(terms), redirect_stdout=True)
        
        # upload Metadata
        for i, term in enumerate(terms):
            if not term in self.__hobo_gid_map.keys():
                print('Term %d has no mapped table in %s' % (term, self._hobo_table_url))
                continue

            # build the url
            url = self._hobo_table_url + self.__hobo_gid_map[term]

            # upload
            cli.metadata(url=url, term=term)
            if not quiet:
                bar.update(i + 1)

    def _download_data(self):
        # download the zip into memory
        r = requests.get(self._hobo_data_url)
        z = zipfile.ZipFile(io.BytesIO(r.content))

        # save to current directory
        z.extractall()
    
    def _upload_data(self, path,  terms='all', quiet=True):
        # if all years, are requested, build the list
        if terms == 'all':
            terms = ['WT17', 'WT18', 'WT19', 'WT20']
        if not isinstance(terms, list):
            terms = [terms]

        # build the cli
        cli = HydenvHoboImporter(connection=self.__connection)

        for term in terms:
            p = os.path.join(path, self.__hobo_data_map[term])
            cli.folder(path=p, term=term, quiet=quiet)



from hydenv.examples.hobo import HydenvHoboExamples
from hydenv.examples.space import HydenvSpaceExamples
from hydenv.examples.customers import HydenvCustomerExamples


class HydenvExamples:
    r"""
    Example Loader.\n
    Loads example files from Github (or other sources) and loads them into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", quiet=False):
        # substitute and save
        self.__connection = connection
        self.quiet = quiet

    def hobo(self, terms='all', only=None, data_path='download'):
        """
        Import HOBO data.\n
        This high level script downloads all neccessary resources for the HOBO 
        measureing campaigns and uploads them to the specified database.
        You can limit the metadata by terms (like WT17, WT18 etc.) and upload only
        'metadata' or 'data'. Default is to upload both.
        :param terms: Either 'all' (default) or a specific term short name (like WT18) to use
        :param data_path: Either 'download' (default) or a local path to the data
        :param only: If set, only the given entities will be uploaded.
        """
        cli = HydenvHoboExamples(connection=self.__connection)
        return cli.run(terms=terms, data_path=data_path, only=only, quiet=self.quiet)

    def space(self, normalize=False):
        """
        Import Space Missions example data.\n
        This high level script downloads a database of all space Mission from 
        1957 until early 2020. The raw, de-normalized dataset is installed along 
        with the hydenv database and used in lecture to introduce SQL and normalization.
        The final normalized structure can be added using the normalize flag.
        :param normalize: If set, the data will be normalized before uplaod.
        """
        cli = HydenvSpaceExamples(connection=self.__connection)
        if normalize:
            raise NotImplementedError
        else:
            return cli.run(quiet=self.quiet)

    def customers(self, k=500, normalize=False):
        """
        Import Customers example data.\n
        This high level script create a set of k random fake customer purchases 
        in fake stores. The data is not normalized and can be used for the 
        normalization sessions.
        :param k: number of purchases to create
        :param normalize: If True, the normalized scheme will be added as well.
        """
        cli = HydenvCustomerExamples(connection=self.__connection)
        return cli.run(quiet=self.quiet, k=k, normalize=normalize)

    def customer(self, **kwargs):
        return self.customers(**kwargs)


if __name__=='__main__':
    import fire
    fire.Fire(HydenvExamples)
